#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.zmi.search module

"""

from elasticsearch_dsl import Q, Search
from pyramid.decorator import reify
from pyramid_es import get_client
from z3c.form.browser.checkbox import SingleCheckBoxFieldWidget
from zope.dublincore.interfaces import IZopeDublinCore
from zope.intid import IIntIds
from zope.schema import Bool, TextLine

from pyams_catalog.query import CatalogResultSet
from pyams_content.shared.common import CONTENT_TYPES, IBaseSharedTool
from pyams_content.shared.common.zmi.search import ISharedToolAdvancedSearchFields, \
    SharedToolAdvancedSearchForm, SharedToolAdvancedSearchResultsView, \
    SharedToolAdvancedSearchView
from pyams_form.interfaces.form import ISearchFormFactory
from pyams_sequence.interfaces import ISequentialIntIds
from pyams_sequence.reference import get_last_version
from pyams_skin.interfaces import IContentSearch
from pyams_skin.layer import IPyAMSLayer
from pyams_template.template import template_config
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.list import unique
from pyams_utils.registry import get_utility
from pyams_workflow.interfaces import IWorkflowVersions


__docformat__ = 'restructuredtext'

from pyams_content_es import _


class IEsSharedToolAdvancedSearchFields(ISharedToolAdvancedSearchFields):
    """Elasticsearch based shared tool advanced search fields"""

    query = TextLine(title=_("Search query"),
                     description=_("This query will applies only on title, short name, header "
                                   "and description; if fulltext search option is selected, this "
                                   "will activate search on all paragraphs, including "
                                   "attachments"),
                     required=False)

    fulltext = Bool(title=_("Fulltext search"),
                    description=_("Search terms in fulltext mode instead of only on contents "
                                  "titles"),
                    required=False,
                    default=False)


@template_config(template='templates/advanced-search.pt', layer=IPyAMSLayer)
class EsSharedToolAdvancedSearchForm(SharedToolAdvancedSearchForm):
    """Elasticsearch shared tool advanced search form"""

    fields_interface = IEsSharedToolAdvancedSearchFields

    @property
    def fields(self):
        fields = super(EsSharedToolAdvancedSearchForm, self).fields
        fields['fulltext'].widgetFactory = SingleCheckBoxFieldWidget
        return fields

    @reify
    def sort_results(self):
        return not bool(self.request.params.get('search_form.widgets.query'))


@adapter_config(context=(IBaseSharedTool, IPyAMSLayer, SharedToolAdvancedSearchView),
                provides=ISearchFormFactory)
@adapter_config(context=(IBaseSharedTool, IPyAMSLayer, SharedToolAdvancedSearchResultsView),
                provides=ISearchFormFactory)
def shared_tool_advanced_search_form_factory(context, request, view):
    """Elasticsearch advanced search form factory"""
    return EsSharedToolAdvancedSearchForm(context, request)


@adapter_config(context=(IBaseSharedTool, IPyAMSLayer, EsSharedToolAdvancedSearchForm),
                provides=IContentSearch)
class EsSharedToolAdvancedSearchFormSearchAdapter(ContextRequestViewAdapter):
    """Elasticsearch shared tool advanced search form adapter"""

    def get_search_results(self, data):
        intids = get_utility(IIntIds)
        params = Q('term', parent_ids=intids.register(self.context)) & \
            Q('terms', content_type=list(CONTENT_TYPES.keys()))
        # check query settings
        query = data.get('query')
        if query:
            sequence = get_utility(ISequentialIntIds)
            if query.startswith('+'):
                params &= Q('term', reference_id=sequence.get_full_oid(query))
            else:
                fulltext = data.get('fulltext', False)
                if fulltext:
                    params &= Q('simple_query_string',
                                query=query)
                else:
                    params &= Q('simple_query_string',
                                query=query,
                                fields=['title.*', 'short_name.*', 'header.*', 'description.*'],
                                default_operator='and')
        if data.get('owner'):
            params &= Q('term', owner_id=data['owner'])
        if data.get('status'):
            params &= Q('term', workflow__status=data['status'])
        if data.get('data_type'):
            params &= Q('term', data_type=data['data_type'])
        if data.get('created_after'):
            params &= Q('range', workflow__created_date={'gte': data['created_after']})
        if data.get('created_before'):
            params &= Q('range', workflow__created_date={'lte': data['created_before']})
        if data.get('modified_after'):
            params &= Q('range', workflow__modified_date={'gte': data['modified_after']})
        if data.get('modified_before'):
            params &= Q('range', workflow__modified_date={'lte': data['modified_before']})
        if data.get('tags'):
            tags = [intids.register(term) for term in data['tags']]
            params &= Q('terms', tags=tags)
        if data.get('themes'):
            themes = [intids.register(term) for term in data['themes']]
            params &= Q('terms', themes__terms=themes)
        if data.get('collections'):
            collections = [intids.register(collection) for collection in data['collections']]
            params &= Q('terms', collections=collections)
        # get ES client and create search object
        client = get_client(self.request)
        search = Search(using=client.es, index=client.index) \
            .params(request_timeout=30) \
            .query(params) \
            .source(['internal_id'])
        sort_values = [{
            'workflow.modified_date': {
                'unmapped_type': 'date',
                'order': 'desc'
            }
        }]
        if query:
            sort_values.insert(0, {
                '_score': {
                    'order': 'desc'
                }
            })
        search = search.sort(*sort_values)[:999]
        # extract results
        if data.get('status'):
            # if status is specified, only extract last version in given status
            return unique(map(lambda x: sorted(IWorkflowVersions(x).get_versions(data['status']),
                                               key=lambda y: IZopeDublinCore(y).modified)[0],
                              CatalogResultSet([result.internal_id for result in search])))
        else:
            # otherwise, extract last absolute version
            return unique(map(get_last_version,
                              CatalogResultSet([result.internal_id for result in search])))
